# MIT License
#
# Copyright (c) 2018 Alexey Nikitin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# todo #14:30m Disable pylint skippinf for this file
# pylint: skip-file
import collections
from datetime import date
from datetime import datetime
from datetime import timedelta
from datetime import tzinfo

import dateutil
from icalendar.cal import Component
from intervaltree import IntervalTree, Interval
from typing import List

from .Intervals import Intervals


class SimpleIntervals(Intervals):

    def __init__(self, tzlocal: tzinfo):
        self.tzlocal = tzlocal
        self.tree = IntervalTree()

    def clear(self):
        self.tree.clear()

    def add(self, components: List[Component], start: datetime, end: datetime):
        for component in components:
            if component.name == 'VEVENT':
                raw_dtstart = component['dtstart']
                dtstart = self.__without_tzinfo(raw_dtstart.dt)
                dtend = self.__without_tzinfo(component.get('dtend', raw_dtstart).dt)
                if self.__is_intersect(start, end, dtstart, dtend):
                    if 'rrule' in component:
                        rrule = dateutil.rrule.rrulestr(
                            component['RRULE'].to_ical().decode('utf-8'),
                            dtstart=dtstart
                        )
                        rrule._until = end
                        duration = dtend - dtstart
                        excludes = set()
                        if 'exdate' in component:
                            if isinstance(component['exdate'], collections.Iterable):
                                for exdates in component['exdate']:
                                    for exdate in exdates.dts:
                                        excludes.add(self.__without_tzinfo(exdate.dt).date())
                            else:
                                for exdate in component['exdate'].dts:
                                    excludes.add(self.__without_tzinfo(exdate.dt).date())
                        for time in filter(lambda time: time.date() not in excludes, list(rrule)):
                            self.tree.add(self.__interval(time, time + duration, component))
                    else:
                        self.tree.add(self.__interval(dtstart, dtend, component))

    def is_inside(self, time: datetime) -> bool:
        return len(self.tree[time]) > 0

    def __interval(self, start: datetime, end: datetime, data: object) -> Interval:
        if start == end:
            end = end + timedelta(seconds=1)
        return Interval(start, end, data)

    def __is_intersect(self, a: datetime, b: datetime, c: datetime, d: datetime) -> bool:
        return max(a, c) <= min(b, d)

    def __without_tzinfo(self, dt):
        if type(dt) == date:
            dt = datetime.combine(dt, datetime.min.time())
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=dateutil.tz.gettz('UTC'))
        return dt.astimezone(self.tzlocal).replace(tzinfo=None)



# MIT License
#
# Copyright (c) 2018 Alexey Nikitin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from meerk.slack import Api

from .Status import Status
from .SyncException import SyncException


class SlackStatus(Status):

    def __init__(self, slack: Api, text: str, emoji: str):
        self.slack = slack
        self.text = text
        self.emoji = emoji

    def sync(self):
        result = self.slack.call(
            'users.profile.set',
            profile={
                'status_text': self.text,
                'status_emoji': self.emoji
            }
        )
        if not result['ok']:
            raise SyncException("Can not sync slack status, because {0}".format(result['error']))



# MIT License
#
# Copyright (c) 2018 Alexey Nikitin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from datetime import datetime
from unittest import TestCase

import dateutil
from icalendar.cal import Component

from meerk.intervals import SimpleIntervals


class SimpleIntervalsTest(TestCase):
    def test_exdates(self):
        components = Component.from_ical(
            """
BEGIN:VEVENT
SUMMARY:asd
DTSTART;TZID=Asia/Barnaul:20180916T230000
DTEND;TZID=Asia/Barnaul:20180916T235900
DTSTAMP:20180917T165114Z
UID:P6nNfouJyandex.ru
SEQUENCE:2
RRULE:FREQ=DAILY;INTERVAL=1
EXDATE;TZID=Asia/Barnaul:20180917T230000
EXDATE;TZID=Asia/Barnaul:20180916T230000
CATEGORIES:Events
CREATED:20180917T162723Z
LAST-MODIFIED:20180917T162741Z
TRANSP:OPAQUE
URL:https://calendar.yandex.ru/for/siberian.pro/event?event_id=673628779
END:VEVENT
            """
        ).walk()
        intervals = SimpleIntervals(dateutil.tz.gettz('UTC'))
        intervals.add(components, datetime(2018, 9, 15, 0, 0), datetime(2018, 9, 17, 0, 0))
        self.assertFalse(intervals.is_inside(datetime(2018, 9, 16, 23, 0)))

    def test_dtstart_equals_dtend(self):
        components = Component.from_ical(
            """
BEGIN:VEVENT
SUMMARY:empty
DTSTART;TZID=Asia/Barnaul:20180927T223000
DTEND;TZID=Asia/Barnaul:20180927T223000
DTSTAMP:20180927T163634Z
UID:ZpNblVw8yandex.ru
SEQUENCE:0
CATEGORIES:Events
CREATED:20180927T163531Z
LAST-MODIFIED:20180927T163531Z
TRANSP:OPAQUE
URL:https://calendar.yandex.ru/for/siberian.pro/event?event_id=685344091
END:VEVENT
            """
        ).walk()
        intervals = SimpleIntervals(dateutil.tz.gettz('UTC'))
        intervals.add(components, datetime(2018, 9, 26, 0, 0), datetime(2018, 9, 28, 0, 0))
        self.assertTrue(intervals.is_inside(datetime(2018, 9, 27, 15, 30)))

    def test_dtend_is_hidden(self):
        components = Component.from_ical(
            """
BEGIN:VEVENT
SUMMARY:Simple
DTSTART:20180926T170000Z
DTSTAMP:20180927T164741Z
UID:someuid
SEQUENCE:0
CREATED:20180927T164659Z
DESCRIPTION:
LAST-MODIFIED:20180927T164716Z
LOCATION:
STATUS:CONFIRMED
TRANSP:OPAQUE
END:VEVENT
            """
        ).walk()
        intervals = SimpleIntervals(dateutil.tz.gettz('UTC'))
        intervals.add(components, datetime(2018, 9, 26, 0, 0), datetime(2018, 9, 28, 0, 0))
        self.assertTrue(intervals.is_inside(datetime(2018, 9, 26, 17, 00)))

    def test_all_day_event(self):
        components = Component.from_ical(
            """
BEGIN:VEVENT
SUMMARY:all day
DTSTART;VALUE=DATE:20181007
DTEND;VALUE=DATE:20181008
DTSTAMP:20181007T043134Z
UID:iq6zz5Cdyandex.ru
SEQUENCE:0
CATEGORIES:events
CREATED:20181007T043121Z
LAST-MODIFIED:20181007T043121Z
TRANSP:OPAQUE
URL:https://calendar.yandex.ru/for/siberian.pro/event?event_id=696510765
END:VEVENT
            """
        ).walk()
        intervals = SimpleIntervals(dateutil.tz.gettz('UTC'))
        intervals.add(components, datetime(2018, 10, 6, 0, 0), datetime(2018, 10, 8, 0, 0))
        self.assertTrue(intervals.is_inside(datetime(2018, 10, 7, 17, 00)))

    def test_all_day_daily_event(self):
        components = Component.from_ical(
            """
BEGIN:VEVENT
SUMMARY:all day
DTSTART;VALUE=DATE:20181007
DTEND;VALUE=DATE:20181008
DTSTAMP:20181007T044204Z
UID:6uk40snnrfsarji6gitijk4k67@google.com
SEQUENCE:0
RRULE:FREQ=DAILY
CREATED:20181007T044142Z
DESCRIPTION:
LAST-MODIFIED:20181007T044142Z
LOCATION:
STATUS:CONFIRMED
TRANSP:TRANSPARENT
END:VEVENT
            """
        ).walk()
        intervals = SimpleIntervals(dateutil.tz.gettz('UTC'))
        intervals.add(components, datetime(2018, 10, 6, 0, 0), datetime(2018, 10, 10, 0, 0))
        self.assertTrue(intervals.is_inside(datetime(2018, 10, 9, 17, 00)))

    def test_all_day_daily_with_modification_event(self):
        components = Component.from_ical(
            """
BEGIN:VEVENT
SUMMARY:all day
DTSTART;VALUE=DATE:20181006
DTEND;VALUE=DATE:20181007
DTSTAMP:20181007T044526Z
UID:0q9s66jci35g6fmjq9bi3m8pbk@google.com
SEQUENCE:0
RRULE:FREQ=DAILY
EXDATE;VALUE=DATE:20181006
CREATED:20181007T044447Z
DESCRIPTION:
LAST-MODIFIED:20181007T044447Z
LOCATION:
STATUS:CONFIRMED
TRANSP:TRANSPARENT
END:VEVENT
            """
        ).walk()
        intervals = SimpleIntervals(dateutil.tz.gettz('UTC'))
        intervals.add(components, datetime(2018, 10, 6, 0, 0), datetime(2018, 10, 10, 0, 0))
        self.assertFalse(intervals.is_inside(datetime(2018, 10, 6, 17, 00)))



# MIT License
#
# Copyright (c) 2018 Alexey Nikitin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from unittest import TestCase

import pycodestyle
from pylint import epylint as lint


# todo #14:30 Add mypy
class StaticAnalysisTest(TestCase):

    def test_style(self):
        result = self.__pycodestyle('../meerk/')
        self.assertEqual(result.total_errors, 0, "Found code style errors or warnings.")

    def test_style_tests(self):
        result = self.__pycodestyle('../tests/')
        self.assertEqual(result.total_errors, 0, "Found tests style errors or warnings.")

    def test_code_lint(self):
        self.__pylint('../meerk')

    def test_tests_lint(self):
        self.__pylint('../tests')

    def __pycodestyle(self, path: str):
        # pylint: disable=R0201
        style = pycodestyle.StyleGuide(config_file='../setup.cfg')
        return style.check_files([path])

    def __pylint(self, path: str):
        (out, err) = lint.py_run(path, return_std=True)
        errors = err.read()  # type: str
        report = out.read()  # type: str
        if errors:
            self.fail(errors)
        elif 'Your code has been rated at 10.00/10' not in report:
            self.fail(report)



# MIT License
#
# Copyright (c) 2018 Alexey Nikitin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import urllib.request
from datetime import datetime
from datetime import timedelta

from icalendar import Calendar as iCalendar

from meerk.intervals import Intervals
from .Calendar import Calendar


class IcsCalendar(Calendar):

    def __init__(self, url: str, intervals: Intervals):
        self.url = url
        self.intervals = intervals

    def is_busy(self, time: datetime) -> bool:
        return self.intervals.is_inside(time)

    def sync(self):
        response = urllib.request.urlopen(self.url)
        ical = response.read()
        self.intervals.clear()
        self.intervals.add(
            iCalendar.from_ical(ical).walk(),
            datetime.now() - timedelta(days=1),
            datetime.now() + timedelta(days=1)
        )
